"""Module to get the position of the Galactic bar of the Milky Way."""
import numpy as np
from astropy.coordinates import Galactocentric
from astropy import units as u
from typing import Optional

def getGalacticBar(length:Optional[float]=6., angle:Optional[float]=-50) -> Galactocentric:
    """
    Get the position of the Galactic bar of the Milky Way.

    Parameters
    ----------
    length:float
        Length of the bar in kpc.
    angle:float
        Angle of the bar in degrees.

    Returns
    -------
    bar:astropy.coordinates.Galactocentric
        The position of the bar.
    """
    x = np.linspace(-length/2., length/2., int(length*10)) * np.cos(np.radians(angle)) * u.kpc
    y = np.linspace(-length/2., length/2., int(length*10)) * np.sin(np.radians(angle)) * u.kpc
    return Galactocentric(x=x, y=y, z=np.zeros(len(x)) * u.kpc,
                          galcen_distance=8.3*u.kpc, z_sun=27*u.pc)





"""Module to work on subcubes."""
from typing import Optional
import astropy.units as u
from spectral_cube import SpectralCube


def extractCutout(cube: SpectralCube,
                  glon: float, glat: float,
                  width: float,
                  v_min: Optional[float] = None, v_max: Optional[float] = None) -> SpectralCube:
    """
    Extract a square cutout from a larger cube.

    Parameters
    ----------
    cube:
        The original SpectralCube to extract from.
    glon:
        Galactic longitude of the center.
    glat:
        Galactic latitude of the center.
    width:
        Width in degrees.
    v_min:
        Minimum velocity in km/s (default = -49.5 km/s).
    v_max:
        Maxiumum velocity in km/s (default = 150 km/s).

    Returns
    -------
    subcube: spectral_cube.SpectralCube
        Returns the extracted subcube.
    """
    glon_min = glon - width / 2
    glon_max = glon + width / 2
    glat_min = glat - width / 2
    glat_max = glat + width / 2

    if v_min is None:
        v_min = -49.5  # [km/s]
    if v_max is None:
        v_max = 150  # [km/s]

    return cube.subcube(xlo=glon_min * u.deg, xhi=glon_max * u.deg,
                        ylo=glat_min * u.deg, yhi=glat_max * u.deg,
                        zlo=v_min * u.km / u.s, zhi=v_max * u.km / u.s)



"""Module for the spiral arms of the Milky Way."""

import numpy as np
from astropy.coordinates import SkyCoord
from astropy.coordinates import Galactocentric
from astropy import units as u
from typing import Optional


def _getSpiralParameters(model):
    # Spiral arm params from Reid et al. (2014)
    reid2014 = {'perseus': {'R_ref': 9.9,  # kpc
                            'beta_ref': 14.2,  # deg (from Galactic center to sun =0)
                            'beta_min': -160.0,  # deg
                            'beta_max': 390.0,  # deg
                            'pitch': 9.9,  # deg
                            'width': 0.38},  # kpc
                'sagittarius': {'R_ref': 6.6,  # kpc
                                'beta_ref': 25.6,  # deg (from Galactic center to sun =0)
                                'beta_min': -245.0,  # deg
                                'beta_max': 225.0,  # deg
                                'pitch': 6.9,  # deg
                                'width': 0.26},  # kpc
                'scutum': {'R_ref': 5.0,  # kpc
                           'beta_ref': 27.6,  # deg (from Galactic center to sun =0)
                           'beta_min': 3.0,  # deg
                           'beta_max': 560.0,  # deg
                           'pitch': 19.8,  # deg
                           'width': 0.17},  # kpc
                'local': {'R_ref': 8.4,  # kpc
                          'beta_ref': 8.9,  # deg (from Galactic center to sun =0)
                          'beta_min': -110.0,  # deg
                          'beta_max': 270.0,  # deg
                          'pitch': 12.8,  # deg
                          'width': 0.33},  # kpc
                'outer': {'R_ref': 13.0,  # kpc
                          'beta_ref': 18.6,  # deg (from Galactic center to sun =0)
                          'beta_min': -70.0,  # deg
                          'beta_max': 410.0,  # deg
                          'pitch': 13.8,  # deg
                          'width': 0.63}}  # kpc


    # Spiral arm params combined from Reid et al. (2019), CK
    custom2019 = {
        'norma1': {'R_ref': 4.46,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': -1.,  # deg
                   'width': 0.14},  # kpc
        'norma0': {'R_ref': 4.46,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': 19.5,  # deg
                   'width': 0.14},  # kpc
        'scutum1': {'R_ref': 4.91,  # kpc
                    'beta_ref': 23.,  # deg (from Galactic center to sun =0)
                    'pitch': 14.1,  # deg
                    'width': 0.23},  # kpc
        'scutum1ck': {'R_ref': 4.91,  # kpc
                      'beta_ref': 23.,  # deg (from Galactic center to sun =0)
                      'pitch': 11.1,  # deg
                      'width': 0.23},  # kpc
        'scutum0': {'R_ref': 4.91,  # kpc
                    'beta_ref': 23.,  # deg (from Galactic center to sun =0)
                    'pitch': 12.1,  # deg
                    'width': 0.23},  # kpc
        'sagittarius1': {'R_ref': 6.04,  # kpc
                         'beta_ref': 24.0,  # deg (from Galactic center to sun =0)
                         'pitch': 17.1,  # deg
                         'width': 0.27},  # kpc
        'sagittarius1ck': {'R_ref': 6.9,  # kpc
                           'beta_ref': 24.0,  # deg (from Galactic center to sun =0)
                           'pitch': 10.,  # deg
                           'width': 0.27},  # kpc
        'sagittarius0': {'R_ref': 6.04,  # kpc
                         'beta_ref': 24.0,  # deg (from Galactic center to sun =0)
                         'pitch': 1.0,  # deg
                         'width': 0.27},  # kpc
        'sagittarius0ck': {'R_ref': 6.04,  # kpc
                           'beta_ref': 24.0,  # deg (from Galactic center to sun =0)
                           'pitch': 7.0,  # deg
                           'width': 0.27},  # kpc
        'perseus1': {'R_ref': 8.87,  # kpc
                     'beta_ref': 40.0,  # deg (from Galactic center to sun =0)
                     'pitch': 10.3,  # deg
                     'width': 0.35},  # kpc
        'perseus0': {'R_ref': 8.87,  # kpc
                     'beta_ref': 40.0,  # deg (from Galactic center to sun =0)
                     'pitch': 8.7,  # deg
                     'width': 0.35},  # kpc
        'perseus0ck': {'R_ref': 8.87,  # kpc
                       'beta_ref': 40.0,  # deg (from Galactic center to sun =0)
                       'pitch': 12.7,  # deg
                       'width': 0.35},  # kpc
        'local': {'R_ref': 8.26,  # kpc
                  'beta_ref': 9.,  # deg (from Galactic center to sun =0)
                  'pitch': 11.4,  # deg
                  'width': 0.31},  # kpc
        'outer1': {'R_ref': 12.24,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': 3.0,  # deg
                   'width': 0.65},  # kpc
        'outer1ck': {'R_ref': 12.24,  # kpc
                     'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                     'pitch': 5.0,  # deg
                     'width': 0.65},  # kpc
        'outer0': {'R_ref': 12.24,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': 9.4,  # deg
                   'width': 0.65},  # kpc
        'outer0ck': {'R_ref': 12.24,  # kpc
                     'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                     'pitch': 11.4,  # deg
                     'width': 0.65}}  # kpc

    # Spiral arm params from Reid et al. (2019)
    # TODO: crosscheck with paper and add correct beta bounds
    reid2019 = {
        'norma1': {'R_ref': 4.46,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': -1.,  # deg
                   'width': 0.14},  # kpc
        'norma0': {'R_ref': 4.46,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': 19.5,  # deg
                   'width': 0.14},  # kpc
        'scutum1': {'R_ref': 4.91,  # kpc
                    'beta_ref': 23.,  # deg (from Galactic center to sun =0)
                    'pitch': 14.1,  # deg
                    'width': 0.23},  # kpc
        'scutum0': {'R_ref': 4.91,  # kpc
                    'beta_ref': 23.,  # deg (from Galactic center to sun =0)
                    'pitch': 12.1,  # deg
                    'width': 0.23},  # kpc
        'sagittarius1': {'R_ref': 6.04,  # kpc
                         'beta_ref': 24.0,  # deg (from Galactic center to sun =0)
                         'pitch': 17.1,  # deg
                         'width': 0.27},  # kpc
        'sagittarius0': {'R_ref': 6.04,  # kpc
                         'beta_ref': 24.0,  # deg (from Galactic center to sun =0)
                         'pitch': 1.0,  # deg
                         'width': 0.27},  # kpc
        'perseus1': {'R_ref': 8.87,  # kpc
                     'beta_ref': 40.0,  # deg (from Galactic center to sun =0)
                     'pitch': 10.3,  # deg
                     'width': 0.35},  # kpc
        'perseus0': {'R_ref': 8.87,  # kpc
                     'beta_ref': 40.0,  # deg (from Galactic center to sun =0)
                     'pitch': 8.7,  # deg
                     'width': 0.35},  # kpc
        'local': {'R_ref': 8.26,  # kpc
                  'beta_ref': 9.,  # deg (from Galactic center to sun =0)
                  'pitch': 11.4,  # deg
                  'width': 0.31},  # kpc
        'outer1': {'R_ref': 12.24,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': 3.0,  # deg
                   'width': 0.65},  # kpc
        'outer0': {'R_ref': 12.24,  # kpc
                   'beta_ref': 18.,  # deg (from Galactic center to sun =0)
                   'pitch': 9.4,  # deg
                   'width': 0.65}}  # kpc

    ck2021 = {'perseus': {'R_ref': 10.9,  # kpc
                      'beta_ref': -16.6,  # deg (from Galactic center to sun =0)
                      'beta_min': -91.415,  # deg (i.e. ~l=280)
                      'beta_max': 0.0,  # deg (i.e. ~l=180)
                      'pitch': 5.7,  # deg
                      'width': 0.38},  # kpc
          'local': {'R_ref': 9.8,  # kpc
                    'beta_ref': -0.5,  # deg (from Galactic center to sun =0)
                    'beta_min': -91.415,  # deg (i.e. ~l=280)
                    'beta_max': 0.0,  # deg (i.e. ~l=180)
                    'pitch': -24,  # deg
                    'width': 0.33},  # kpc
          'outer': {'R_ref': 14.4,  # kpc
                    'beta_ref': 180.,  # deg (from Galactic center to sun =0)
                    'beta_min': -91.415,  # deg (i.e. ~l=280)
                    'beta_max': 0.0,  # deg (i.e. ~l=180)
                    'pitch': 15.8,  # deg
                    'width': 0.63}}  # kpc

    xinyu2016 = {'outer': {'R_ref': 13.6,  # kpc
                           'beta_ref': 26.9,  # deg (from Galactic center to sun =0)
                           'beta_min': -110.0,  # deg
                           'beta_max': 270.0,  # deg
                           'pitch': 13.1,  # deg
                           'width': np.nan}}  # kpc

    vallee2015 = {'perseus': {'R_ref': 7.0,  # kpc
                              'beta_ref': 90.,  # deg (from Galactic center to sun =0)
                              'beta_min': -160.0,  # deg
                              'beta_max': 390.0,  # deg
                              'pitch': 13.,  # deg
                              'width': 0.38},  # kpc
                  'sagittarius': {'R_ref': 7.0,  # kpc
                                  'beta_ref': 0.,  # deg (from Galactic center to sun =0)
                                  'beta_min': -245.0,  # deg
                                  'beta_max': 225.0,  # deg
                                  'pitch': 13.,  # deg
                                  'width': 0.33},  # kpc
                  'scutum': {'R_ref': 7.0,  # kpc
                             'beta_ref': 270.,  # deg (from Galactic center to sun =0)
                             'beta_min': 3.0,  # deg
                             'beta_max': 560.,  # deg
                             'pitch': 13.,  # deg
                             'width': 0.33},  # kpc
                  'outer': {'R_ref': 7.0,  # kpc
                            'beta_ref': 180.,  # deg (from Galactic center to sun =0)
                            'beta_min': -70.0,  # deg
                            'beta_max': 410.0,  # deg
                            'pitch': 12.5,  # deg
                            'width': 0.63}}  # kpc

    if model == 'reid2014':
        spiral_arm_params = reid2014
    elif model == 'reid2019':
        spiral_arm_params = reid2019
    elif model == 'xinyu2016':
        spiral_arm_params = xinyu2016
    elif model == 'vallee2015':
        spiral_arm_params = vallee2015
    elif model == 'ck2021':
        spiral_arm_params = ck2021

    return spiral_arm_params


def getSpiralArm(name:str, model:str='reid2014',
                 glon_min:Optional[float]=-np.inf, glon_max:Optional[float]=np.inf,
                 r_gal_min:Optional[float]=0., r_gal_max:Optional[float]=np.inf,
                 resolution:Optional[float]=0.0001, R_0:Optional[float]=8.34) -> Galactocentric:
    """
    Get the spiral arm model for a given arm name.

    Parameters
    ----------
    name: str
        Name of the spiral arm.
    model: str
        Name of the spiral arm model to be used.
        Available models: reid2014 (default), ck, vallee2015.
    glon_min: float
        Minimum galactic longitude [deg] of the spiral arm in degrees.
    glon_max: float
        Maximum galactic longitude [deg] of the spiral arm in degrees.
    r_gal_min: float
        Minimum galactocentric distance [kpc] of the spiral arm in kpc.
    r_gal_max: float
        Maximum galactocentric distance [kpc] of the spiral arm in kpc.
    resolution: float
        Resolution of the spiral arm model in degrees.
    R_0: float
        Galactocentric distance of the sun in kpc.

    Returns
    -------
    coordinates: astropy.coordinates.Galactocentric
        Coordinates of the spiral arm in the galactocentric frame.
    """
    spiral_arm_params = _getSpiralParameters(model)

    R_ref = spiral_arm_params[name]['R_ref']
    beta_ref = spiral_arm_params[name]['beta_ref']
    pitch = spiral_arm_params[name]['pitch']
    beta_min = spiral_arm_params[name]['beta_min']
    beta_max = spiral_arm_params[name]['beta_max']
    beta = np.arange(beta_max, beta_min, -1. * resolution)  # [deg]

    R_gal = R_ref * np.exp(-1 * (beta - beta_ref) * np.pi / 180. * np.tan(np.pi / 180. * pitch))  # kpc

    #     x = R_gal * -1. * np.cos(beta * np.pi/180.)
    #     y = R_gal * np.sin(beta * np.pi/180.)
    x = -1 * R_gal * np.sin((90 - beta) * np.pi / 180.)
    y = R_gal * np.cos((90 - beta) * np.pi / 180.)

    arm = SkyCoord(frame='galactocentric', x=x*u.kpc, y=y*u.kpc, z=0.0*u.kpc,
                   galcen_distance=R_0*u.kpc,
                   galcen_coord=SkyCoord(frame='galactic', l=0*u.deg, b=0*u.deg).icrs,
                   z_sun=0.*u.kpc,
                   roll=0.).galactocentric
    arm = arm[arm.galactic.l.degree > glon_min]
    arm = arm[arm.galactic.l.degree < glon_max]
    arm = arm[arm.galactic.distance.kpc > r_gal_min]
    arm = arm[arm.galactic.distance.kpc < r_gal_max]
    return arm


def getSpiralArmsDetail(model='reid2019'):

    # if model=='optimized':
    #     # SAGITTARIUS arm
    #     svx1, svy1 = getSpiralArm('sagittarius', beta_min=-245., beta_max=-2, model='vallee2015')
    #     sx, sy = getSpiralArm('sagittarius', beta_min=-2, beta_max=68, model='reid2014')
    #     svx2, svy2 = getSpiralArm('sagittarius', beta_min=68, beta_max=225, model='vallee2015')
    #
    #     # SCUTUM arm
    #     scx, scy = getSpiralArm('scutum', beta_min=3, beta_max=101)
    #     scvx, scvy = getSpiralArm('scutum', beta_min=35, beta_max=363, model='vallee2015')
    #     scvx2, scvy2 = getSpiralArm('scutum', beta_min=461, beta_max=560, model='vallee2015')
    #
    #     # OUTER arm
    #     ox, oy = getSpiralArm('outer', beta_min=-6, beta_max=56)
    #     ovx, ovy = getSpiralArm('outer', beta_min=56, beta_max=410, model='vallee2015')
    #     ockx, ocky = getSpiralArm('outer', beta_min=-70, beta_max=-6, model='ck')
    #
    #     # PERSEUS spiral arm
    #     px, py = getSpiralArm('perseus', beta_min=-21, beta_max=88)
    #     pvx, pvy = getSpiralArm('perseus', beta_min=90, beta_max=390, model='vallee2015')
    #     pckx, pcky = getSpiralArm('perseus', beta_min=-40, beta_max=-10, model='ck')
    #     pvx2, pvy2 = getSpiralArm('perseus', beta_min=-160, beta_max=-40, model='vallee2015')
    #
    #     return sy, sx, svy1, svx1, svy2, svx2, scy, scx, scvy, scvx, scvy2, scvx2, oy, ox, ovy, ovx, ocky, ockx, py, px, pvy, pvx, pvy2, pvx2, pcky, pckx
    # else:

    # SAGITTARIUS arm
    # svx1, svy1 = getSpiralArm('sagittarius', beta_min=-245., beta_max=-2, model=model)
    # sx, sy = getSpiralArm('sagittarius', beta_min=-2, beta_max=68, model=model)
    # svx2, svy2 = getSpiralArm('sagittarius', beta_min=68, beta_max=225, model=model)
    # sag_x = list(svx1)+list(sx)+list(svx2)
    # sag_y = list(svy1)+list(sy)+list(svy2)
    sagittarius = getSpiralArm('sagittarius', model=model)

    # SCUTUM arm
    # scx, scy = getSpiralArm('scutum', beta_min=3, beta_max=101, model=model)
    # scvx, scvy = getSpiralArm('scutum', beta_min=101, beta_max=363, model=model)
    # scvx2, scvy2 = getSpiralArm('scutum', beta_min=363, beta_max=560, model=model)
    # scutum_x = list(scx)+list(scvx)+list(scvx2)
    # scutum_y = list(scy)+list(scvy)+list(scvy2)
    scutum = getSpiralArm('scutum', model=model)

    # OUTER arm
    # ockx, ocky = getSpiralArm('outer', beta_min=-70, beta_max=-6, model=model)
    # ox, oy = getSpiralArm('outer', beta_min=-6, beta_max=56, model=model)
    # ovx, ovy = getSpiralArm('outer', beta_min=56, beta_max=410, model=model)
    # outer_x = list(ockx) + list(ox) + list(ovx)
    # outer_y = list(ocky) + list(oy) + list(ovy)
    outer = getSpiralArm('outer', model=model)

    # PERSEUS spiral arm
    # pvx2, pvy2 = getSpiralArm('perseus', beta_min=-160, beta_max=-40, model=model)
    # pckx, pcky = getSpiralArm('perseus', beta_min=-40, beta_max=-10, model=model)
    # px, py = getSpiralArm('perseus', beta_min=-10, beta_max=90, model=model)
    # pvx, pvy = getSpiralArm('perseus', beta_min=90, beta_max=390, model=model)
    # per_x = list(pckx) + list(pvx2) + list(px) + list(pvx)
    # per_y = list(pcky) + list(pvy2) + list(py) + list(pvy)
    perseus = getSpiralArm('perseus', model=model)

    return sagittarius, scutum, outer, perseus


def getSpiralArms(model:Optional[str]='reid2014') -> dict:
    """
    Returns the spiral arms in a dictionary.

    Parameters
    ----------
    model: str
        The model to use

    Returns
    -------
    arms: dict
        A dictionary with the spiral arms.
    """
    sagittarius, scutum, outer, perseus = getSpiralArmsDetail(model=model)

    return {'sagittarius': sagittarius,
            'scutum': scutum,
            'perseus': perseus,
            'outer': outer}
    #
    # return {'sagittarius': {'x': list(svx1) + list(sx) + list(svx2),
    #                         'y': list(svy1) + list(sy) + list(svy2)},
    #         'scutum': {'x': list(scvx) + list(scx) + list(scvx2),
    #                    'y': list(scvy) + list(scy) + list(scvy2)},
    #         'perseus': {'x': list(pvx) + list(px) + list(pckx) + list(pvx2),
    #                     'y': list(pvy) + list(py) + list(pcky) + list(pvy2)},
    #         'outer': {'x': list(ovx) + list(ox) + list(ockx),
    #                   'y': list(ovy) + list(oy) + list(ocky)}}



"""Module for coordinate operations in the Milky Way."""

import numpy as np

def _residual_brand(R_gal, vlsr, glon, glat=0., R_0=8.34, t_0=240.):
    '''
    Calculate residual for Brand & Blitz (1993) rotation curve, using parameters from
    Reid 2014.

    @param r: Galactocentric radius [kpc]
    @type r: float
    @param vlsr: Line of sight velocity with respect to local standard of rest [km s^-1]
    @type vlsr: float
    @param glon: Galactic longitude [deg]
    @type glon: float
    @param glat: (optional; default glat=0) Galactic latitude [deg]
    @type glat: float
    @param R_0: (optional; default R_0=8.34) Distance to Galactic center [kpc]
    @type R_0: float
    @param t_0: (optional; default Theta_0=248.) Circular rotation velocity at orbit of
                the sun [km s^-1]
    @type t_0: float
    @return: Residual between the actual vlsr and the expected from the rotation curve
    @rtype: float

    '''
    a1 = 1.00767
    a2 = 0.0394
    a3 = 0.00712
    a = vlsr
    c = np.sin(glon * np.pi / 180.) * np.cos(glat * np.pi / 180.) * t_0 * (
                a1 * (R_gal / R_0) ** (a2 - 1) + a3 * (R_0 / R_gal) - 1)
    return c - a


def getRgalFromVlsr(vlsr, glon, glat=0., R_0=8.34, t_0=240., precision=0.001):
    '''
    Calculate Galactocentric radius from line of sight velocity and Galactic longitude.

    @param vlsr: Line of sight velocity with respect to local standard of rest [km s^-1]
    @type vlsr: float
    @param glon: Galactic longitude [deg]
    @type glon: float
    @param glat: (optional; default glat=0) Galactic latitude [deg]
    @type glat: float
    @param R_0: (optional; default R_0=8.34) Distance to Galactic center [kpc]
    @type R_0: float
    @param t_0: (optional; default Theta_0=248.) Circular rotation velocity at orbit of
                the sun [km s^-1]
    @type t_0: float
    @param precission: (optional; default prec=0.001) Precission of the result [km s^-1]
    @type precission: float
    @return: Galactocentric radius [kpc]
    @rtype: float

    '''
    # make it 2D and column vector (i.e. transpose) if array is given
    if hasattr(vlsr, '__len__'):
        # solution for input arrays
        if type(vlsr) != np.array:
            vlsr = np.array(vlsr)
        vlsr = vlsr[np.newaxis].T
    if hasattr(glon, '__len__'):
        # solution for input arrays
        if type(glon) != np.array:
            glon = np.array(glon)
        glon = glon[np.newaxis].T

    distances = np.arange(0.1, 30. + precision, precision)
    residuals = _residual_brand(distances, vlsr, glon, glat=glat, t_0=t_0, R_0=R_0)
    min_index = np.argmin(np.abs(residuals), axis=len(residuals.shape) - 1)
    R_gal_min = distances[min_index]
    print(distances)

    try:
        # solution for input arrays
        # assume no solution is found if in the extreme values
        is_invalid = (min_index == 0) + (min_index == len(distances) - 1)
        R_gal_min[is_invalid] = np.nan
    except TypeError:
        # solution for single float input
        if min_index in (0, len(distances)):
            R_gal_min = np.nan

    return np.round(R_gal_min, int(np.abs(np.log10(precision))))


def getVlsrFromRgal(R_gal, glon, glat=0., R_0=8.34, t_0=240, precision=0.001):
    '''
    Calculate line of sight velocity from Galactocentric radius.

    @param R_gal: Galactocentric radius [kpc]
    @type R_gal: float
    @param glon: Galactic longitude [deg]
    @type glon: float
    @param glat: (optional; default glat=0) Galactic latitude [deg]
    @type glat: float
    @param R_0: (optional; default R_0=8.34) Distance to Galactic center [kpc]
    @type R_0: float
    @param t_0: (optional; default Theta_0=248.) Circular rotation velocity at orbit of
                the sun [km s^-1]
    @type t_0: float
    @param precission: (optional; default prec=0.001) Precission of the result [km s^-1]
    @type precission: float
    @return: Line of sight velocity with respect to local standard of rest [km s^-1]
    @rtype: float

    '''
    # make it 2D and column vector (i.e. transpose) if array is given
    if hasattr(R_gal, '__len__'):
        # solution for input arrays
        if type(R_gal) != np.array:
            R_gal = np.array(R_gal)
        R_gal = R_gal[np.newaxis].T
    if hasattr(glon, '__len__'):
        # solution for input arrays
        if type(glon) != np.array:
            glon = np.array(glon)
        glon = glon[np.newaxis].T

    vlsrs = np.arange(-200., 200. + precision, precision)
    residuals = _residual_brand(R_gal, vlsrs, glon, glat=glat, t_0=t_0, R_0=R_0)
    min_index = np.argmin(np.abs(residuals), axis=len(residuals.shape) - 1)
    vlsr_min = vlsrs[min_index]

    try:
        # solution for input arrays
        # assume no solution is found if in the extreme values
        is_invalid = (min_index == 0) + (min_index == len(vlsrs) - 1)
        vlsr_min[is_invalid] = np.nan
    except TypeError:
        # solution for single float input
        if vlsr_min in (-200., 200.):
            vlsr_min = np.nan

    return np.round(vlsr_min, int(np.abs(np.log10(precision))))
